﻿using System;
using System.Reflection;
using HeuristicLab.Persistence.Core;

namespace HeuristicLab.Persistence {

  public class DataMemberAccessor {

    public delegate object Getter();
    public delegate void Setter(object value);

    public readonly Getter Get;
    public readonly Setter Set;
    public readonly string Name;
    public readonly Type Type;
    public readonly object DefaultValue;

    public DataMemberAccessor(
        MemberInfo memberInfo,
        StorableAttribute autoStorableAttribute,
        object obj) {
      if (memberInfo.MemberType == MemberTypes.Field) {
        FieldInfo fieldInfo = (FieldInfo)memberInfo;
        Get = () => fieldInfo.GetValue(obj);
        Set = value => fieldInfo.SetValue(obj, value);
        Type = fieldInfo.FieldType;
      } else if (memberInfo.MemberType == MemberTypes.Property) {
        PropertyInfo propertyInfo = (PropertyInfo)memberInfo;
        if (!propertyInfo.CanRead || !propertyInfo.CanWrite) {
          throw new NotSupportedException(
            "Storable properties must implement both a Get and a Set Accessor. ");
        }
        Get = () => propertyInfo.GetValue(obj, null);
        Set = value => propertyInfo.SetValue(obj, value, null);
        Type = propertyInfo.PropertyType;
      } else {
        throw new NotSupportedException(
                "The Storable attribute can only be applied to fields and properties.");
      }
      Name = autoStorableAttribute.Name ?? memberInfo.Name;
      DefaultValue = autoStorableAttribute.DefaultValue;
    }

    public DataMemberAccessor(
        string name, Type type, object defaultValue,
        Getter getter, Setter setter) {
      Name = name;
      Type = type;
      DefaultValue = defaultValue;
      Get = getter;
      Set = setter;
    }

    public DataMemberAccessor(object o) {
      Name = null;
      Type = o.GetType();
      DefaultValue = null;
      Get = () => o;
      Set = null;
    }

    public override string ToString() {
      return String.Format("DataMember({0}, {1}, {2}, {3}, {4})",
        Name,
        Type == null ? "<null>" : Type.FullName,
        DefaultValue ?? "<null>",
        Get.Method, Set.Method);
    }
  }

}